/********************************************************************************

 Programmed by SEUNG-JO BAE 

  - Goal: PACK

  - Algorithm: Seungjo Bae & Sanjay Ranka

  - Input: Input array 'A' & Mask array 'M'

  - Output: Packed vector 'V'

  - Version 1.1 (4/26/95)

  - Note: 1) Not scalable.
             Use different functions for 1-D & 2-D.
 
	  2) Linear permutation using active message-PUT (4/26)

          3) New send_sendl using CMAML_rpc (5/6)

          4) Async. comm. (5/6)

	  5) Linear permutation using CMMD_swap (5/6)

	  6) ManytoMany_Con(sendB, sendl, send_offset, msg_size, 
	                    recvB, recvl, recv_offset, 1, -1, -1) and
	     ManytoMany_Con(sendB, sendl, send_offset, msg_size, 
                            recvB, recvl, recv_offset, 2, -1, -1)
	     work well, so we can say there is no interface problem
             between my program and Khaled's program. (5/9)

	  7) New PRS_dim using CMAML_scopy (5/10)

	  8) New random_mask() (5/10)

	  9) Revise final() for computing Vwidth & V_L. (5/10)

	 10) Check if mask array is false. (5/10)

	 11) New timing method (5/10) (5/11)

	 12) Advanced PRS (Kim's routine) when vector size > nproc (6/28)

	 13) Modify naive prs0 (7/3)

         14) Use dynamic allocation for PS, RS (7/14)

	 15) Khaled's two-stage algorithm finally worked, but distributed
             still did not work (Wrong results) (7/17)
  
	 16) Use static allocation for PS, RS (7/21)

	 17) Delete unnecessary local copy when using SCARAB (7/26)

         18) Modify scount in comp_mesg (8/9)

********************************************************************************/

#include <cm/cmmd.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <cmsys/cm_fcntl.h>

/*****
#include "global.h"
*****/

/*****/
#include "global-1d.h"
/*****/

#include "function.h"

#include "print.h"

#define  MACHINE CM_5
/*#include "/home/R6G/sbae/Khaled/scarab.h"*/

int rport = 8;



/******************************************************************************
 *                                                                            *
 *  Function : read_input()                                                   *
 *                                                                            *
 ******************************************************************************/


void read_input()
{
   Boolean error;
   int i, j;
   int total = 1, t_total=1;
   int save;

   CMMD_disable_safety();
   CMMD_fset_io_mode(stdin, CMMD_sync_bc);
   CMMD_fset_io_mode(stdout, CMMD_sync_bc);
   CMMD_fset_io_mode(stderr, CMMD_independent);

   ori_nproc = CMMD_partition_size();

   /*---------------- Rank of the input Array ---------------------------*/
   do {
/***
      printf("\nEnter rank of the input array (Max=%d): ", Rankmax);
***/
      scanf("%d", &rank);
      error = ( rank < 1 || rank > Rankmax );
      if (error)
	 printf("\n??? Improper rank. Try again!\n");
    } while (error);

   /*-------------------- from dim=0 to dim=rank-1 ----------------------*/
   for (i=0; i<rank; ++i) {

/***
       printf("\n***** Dimension = %d *****", i);
***/

       do {
/***
	 printf("\n   Enter domain size (Max=%d): ", Nmax);
***/
	 scanf("%d", &N[i]);
	 error = ( N[i] < 1 || N[i] > Nmax || !Blog2(N[i]) );
	 if (error)
	   printf("\n??? Improper domain size. Try again!\n");
       } while (error);

       /*------------ Source Distribution for A, M ------------------*/
       save = total;
       do {
/***
	 printf("\n   Enter number of source processors (Max=%d): ", 
                                     Procmax/total);
***/
	 scanf("%d", &nproc[i]);
	 total *= nproc[i];
	 error = ( !Blog2(nproc[i])  || nproc[i] < 1 ||
		   total > ori_nproc || nproc[i] > N[i] );
	 dproc[i] = Ilog2(nproc[i]);
	 if (error) {
	   total = save;
	   printf("\n??? Improper source processor number. Try again!\n");
	 }
       } while (error);
       
       if ( nproc[i] == 1 ) {
	 width[i] = N[i];
	 L[i]     = N[i];
	 if ( L[i] > N_Lmax )
	   CMMD_error("Error: Increas 'N_Lmax' !\n");
       }
       else {
	 do {
/***
	   printf("\n   Enter width of a partition (Max=%d): ", 
                          N[i] / nproc[i]);
***/
	   scanf("%d", &width[i]);
	   L[i] = N[i]/nproc[i];
	   error = ( !Blog2(width[i]) || width[i] < 1 || 
                      width[i] > L[i] || L[i] > N_Lmax );
	   if (error)
	     printf("\n??? Improper source partition width. Try again!\n");
	 } while (error);
       }
     }

   tot_nproc = total;
   tot_dproc = Ilog2(tot_nproc);

   i = CMMD_reset_partition_size(tot_nproc);
   if (i != tot_nproc) 
     CMMD_error("\nERROR: CMMD Partition failed!\n");

   /* every processor use a different random seed */
   /*srandom(CMMD_self_address()); */

   if ( CMAML_allocate_this_rport(rport) <= 0 )
     CMMD_error("ERROR: Failed to allocate receive rport %d\n", rport);

 }


/******************************************************************************
 *                                                                            *
 *  Function : irregular_mask()
 *                                                                            *
 ******************************************************************************/


void irregular_mask1()
{
   int g,      /* global index */
       l,      /* local index  */
       pn;
 
   int m; 

   CMMD_sync_with_nodes(); 

   for (g=0; g<N[0]; ++g) {
     CMMD_sync_with_nodes(); 

     l  = map_to_local(g, 0);
     pn = map_to_proc(g, 0);

     if ( my_pnum == pn ) {
       if ( g < N[0]/2 )
	 M1[l] = 1;
       else
	 M1[l] = 0;
     }
   }
};


/*----------------------------------------------------------*/

void irregular_mask2()
{
   int gi, gj,     /* global index */
       li, lj,     /* local index  */
       pn;
   int gpn[Rankmax];
   int m;
 
   CMMD_sync_with_nodes(); 

   for (gi=0; gi<N[1]; ++gi) {
     CMMD_sync_with_nodes(); 

     for (gj=0; gj<N[0]; ++gj) {  
       CMMD_sync_with_nodes();

       lj = map_to_local(gj, 0);
       li = map_to_local(gi, 1);

       gpn[0] = map_to_proc(gj, 0);
       gpn[1] = map_to_proc(gi, 1);

       map_to_pnum(gpn, &pn);

       if ( my_pnum == pn ) {
	 if (gj < gi ) 
	   M2[li][lj] = 1;
	 else
	   M2[li][lj] = 0;
       }
             
     }; /* for(gj) */

   }; /* for(gi) */

};

/******************************************************************************
 *                                                                            *
 *  Function : read_mask()                                                    *
 *                                                                            *
 ******************************************************************************/


void read_mask1(name)
     char *name;
{
   FILE *fp;
   int g,      /* global index */
       l,      /* local index  */
       pn;
 
   int m; 

   if (! strcmp(name, "ir"))
       irregular_mask1();

   else {
     if ( !(fp = fopen(name, "r")) )
       CMMD_error("\nERROR: Fail to open the file! \n");

     CMMD_sync_with_nodes(); 
     CMMD_fset_io_mode(fp, CMMD_sync_bc); 

     for (g=0; g<N[0]; ++g) {
       CMMD_sync_with_nodes(); 

       l  = map_to_local(g, 0);
       pn = map_to_proc(g, 0);

       fscanf(fp, "%d", &m);

       if ( my_pnum == pn ) 
	 M1[l] = m;

/***
CMMD_sync_with_nodes(); 
CMMD_fset_io_mode(stdout, CMMD_sync_bc); 
printf("\n pn=%d, global=%d, local=%d, mask=%d", pn, g, l, m);
***/
         
     };

     fclose(fp);

   } /* else */

};


/*----------------------------------------------------------*/

void read_mask2(name)
     char *name;
{
   FILE *fp;
   int gi, gj,     /* global index */
       li, lj,     /* local index  */
       pn;
   int gpn[Rankmax];
   int m;
 
   if (! strcmp(name, "ir"))
       irregular_mask2();

   else {
     if ( !(fp = fopen(name, "r")) )
       CMMD_error("\nERROR: fail to open the file!\n");

     CMMD_sync_with_nodes(); 
     CMMD_fset_io_mode(fp, CMMD_sync_bc); 
     for (gi=0; gi<N[1]; ++gi) {
       CMMD_sync_with_nodes(); 

       for (gj=0; gj<N[0]; ++gj) {  
	 CMMD_sync_with_nodes();

	 lj = map_to_local(gj, 0);
	 li = map_to_local(gi, 1);

	 gpn[0] = map_to_proc(gj, 0);
	 gpn[1] = map_to_proc(gi, 1);

	 map_to_pnum(gpn, &pn);

	 fscanf(fp, "%d", &m);

	 if ( my_pnum == pn ) 
	   M2[li][lj] = m;
             
       }; /* for(gj) */

     }; /* for(gi) */

     fclose(fp);
   } /* else */

};
 


/******************************************************************************
 *                                                                            *
 *  Function : random_mask()
 *                                                                            *
 ******************************************************************************/


void random_mask1(density)  /* use only one random number generator */
     int density;
{
  double den;
  int g,      /* global index */
      l,      /* local index  */
      pn;
  Boolean b;
 
  den = (double)density / 100.;

  for (g=0; g<N[0]; ++g) {

    l  = map_to_local(g, 0);
    pn = map_to_proc(g, 0);

    b = drandom() < den;

    if ( my_pnum == pn ) 
      M1[l] = b;

  }

}


void random_mask2(density)  /* use only one random number generator */
     int density;
{

  double den;
  int gi, gj,     /* global index */
      li, lj,     /* local index  */
      pn;
  int gpn[Rankmax];
  Boolean b;
 
  den = (double)density / 100.;

  for (gi=0; gi<N[1]; ++gi) {
    
    for (gj=0; gj<N[0]; ++gj) {  

      b = drandom() < den;

      lj = map_to_local(gj, 0);
      li = map_to_local(gi, 1);

      gpn[0] = map_to_proc(gj, 0);
      gpn[1] = map_to_proc(gi, 1);

      map_to_pnum(gpn, &pn);

      if ( my_pnum == pn )
	M2[li][lj] = b;
             
    }; /* for(gj) */

  }; /* for(gi) */

}

void random_mask1_sep(density)
     int density;
{
  int i;
  double den;

  den = (double)density / 100.;

/***
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n==============>> density = %d, den = %f\n", density, den);
***/

  for (i=0; i<L[0]; ++i) {
    if (drandom() < den)
      M1[i] = true;
    else
      M1[i] = false;
  }

}


void random_mask2_sep(density)
     int density;
{

  int i, j;     /* local index  */
  double den;

  den = (double)density / 100.;

/***
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n==============>> density = %d, den = %f\n", density, den);
***/

  for (i=0; i<L[1]; ++i) {
    for (j=0; j<L[0]; ++j) 
      if (drandom() < den)
	M2[i][j] = true;
      else
	M2[i][j] = false;
  }

}



/******************************************************************************
 *                                                                            *
 *  Function : initialize()                                                   *
 *                                                                            *
 ******************************************************************************/


void initialize()
{
  int i, dim, pn;

   /*======= Who I am ? =======*/
   my_pnum  = CMMD_self_address(); 
   map_to_grid(my_pnum, my_grid);

   /*====== Map-to-Grid ======*/
   for (i=0; i<tot_nproc; ++i) 
     map_to_grid(i, other_grid[i]);

  /*======= Dist info =======*/
  for (dim=0; dim<rank; ++dim) {
    twidth[dim] = nproc[dim]*width[dim];
    ntile[dim]  = L[dim]/width[dim];
  }

  /*======= Clear comm. vectors =======*/
  for (pn=0; pn<tot_nproc; ++pn) {
    sendl[pn] = recvl[pn] = 0;
    send_offset[pn] = recv_offset[pn] = 0;
  }

  CMMD_node_timer_stop(0); 
  CMMD_node_timer_stop(4); 

/********************************* Full (Long) Result

  CMMD_fset_io_mode(stdout,CMMD_sync_bc);
  printf("\n\n**************************************************************");
  printf("\n                  INPUT DATA (Rank = %d)", rank);
  printf("\n**************************************************************");
  printf("\n Total number of processors = %d", tot_nproc);
  printf("\n--------------------------------------------------------------");

  for (dim=0; dim<rank; ++dim) {
    printf("\n <<< Dimension = %d >>>", dim);
    printf("\n Domain size     = %d\n", N[dim]);
    printf("\n Processors      = %d", nproc[dim]);
    printf("\n Partition Width = %d", width[dim]);
    printf("\n Local domain    = %d", L[dim]);
    printf("\n Tile Width      = %d", twidth[dim]);
    printf("\n Number of Tile  = %d", ntile[dim]);
    printf("\n--------------------------------------------------------------");
    
  }
  printf("\n");
  fflush(stdout);
  CMMD_sync_with_nodes();

*********************************/

  CMMD_node_timer_start(0);
  CMMD_node_timer_start(4); 
}



/******************************************************************************
 *                                                                            *
 *  Function : make_array*()                                                  *
 *                                                                            *
 ******************************************************************************/


void make_array1()
{  
  int i, /* local */
      r; /* global */

  for (i=0; i<L[0]; ++i) {
    r = map_to_global(my_pnum, i, 0);
    A1[i] = r;
  }

}


void make_array2()
{
  int r, c,     /* global index */
      i, j;     /* local index  */

  for (i=0; i<L[1]; ++i) {
    r = map_to_global(my_grid[1], i, 1);
    for (j=0; j<L[0]; ++j) {
      c = map_to_global(my_grid[0], j, 0);
      A2[i][j] = r*N[0] + c;
    }
  }

}
  


/******************************************************************************
 *                                                                            *
 *  Function : local_scan();                                                *
 *                                                                            *
 ******************************************************************************/


void local_scan1()
{
  int i,  /* local index */
      t,  /* tile number */
      j;  /* 0 ... Width[0]-1 */
  int lrank;     /* initial rank is valid in each tile */

  PRS_size[0] = ntile[0];

/******************************************
  Allocation for RS0 & PS0 ==> 
    Not work! Change to static array (4/27) 

  RS[0] = (int *)malloc( ntile[0] );
  PS[0] = (int *)malloc( ntile[0] );

  I got it! (7/14).

  RS[0] = (int *)malloc( PRS_size[0] * datasize );
  PS_[0] = (int *)malloc( PRS_size[0] * datasize);

*******************************************/ 

  LPsize = 0;

  /* Local scan and Initialize RS0 & PS0 */
  for (t=0; t<ntile[0]; ++t) {
    lrank = 0;
    for (j=0; j<width[0]; ++j) {
      i = t*width[0] + j;
      if (M1[i]) {
	storage[LPsize].lidx[0] = i;    /* local index */
	storage[LPsize].tnum0 = t;      /* tile number */
	storage[LPsize].rank = lrank;    /* initial rank */
	++LPsize;
        ++lrank;
      }
    }
    RS[0][t] = PS_[0][t] = lrank;
  }

/***** For debug: print out array *****
print_PRS1(PS_[0], 'P');
print_PRS1(RS[0], 'R');
**************************************/

}


/*---------------------------------------------------------------------------*/


#define PS0(r,c)  *(PS[0] + r*ntile[0] + c)
#define RS0(r,c)  *(RS[0] + r*ntile[0] + c)

void local_scan2()
{
  int i0, i1, /* local index */
      t,      /* tile number */
      j;      /* 0 ... Width[0]-1 */
  int count;
  int lrank;   /* initial rank is valid in each tile0 */


  PRS_size[0] = L[1] * ntile[0];
  PRS_size[1] = ntile[1];

/******************************************************  
  Allocation for RS0 & PS0 and RS1 & PS1 ==> 
     Not work! Change to static array (4/27) 

  RS[0] = (int *)malloc( PRS_size[0] );
  PS_[0] = (int *)malloc( PRS_size[0] );

  RS[1] = (int *)malloc( PRS_size[1] );
  PS_[1] = (int *)malloc( PRS_size[1] );

*******************************************************

  I got it! (7/14)

  RS[0] = (int *)malloc( PRS_size[0] * datasize );
  PS_[0] = (int *)malloc( PRS_size[0]  * datasize );

  RS[1] = (int *)malloc( PRS_size[1]  * datasize );
  PS_[1] = (int *)malloc( PRS_size[1] * datasize  );

********************************************************/

  LPsize = 0;

  /* Local scan and Initialize RS0 & PS0 */
  count = 0;
  for (i1=0; i1<L[1]; ++i1) {
    for (t=0; t<ntile[0]; ++t) {

      lrank = 0;
      for (j=0; j<width[0]; ++j) {
	i0 = t*width[0] + j; 
	if (M2[i1][i0]) {
	  storage[LPsize].lidx[0] = i0;    /* local index dim=0 */
	  storage[LPsize].lidx[1] = i1;    /* local index dim=1 */
	  storage[LPsize].tnum0 = t;       /* tile number dim=0 */
	  storage[LPsize].rank = lrank;     /* initial rank */
	  ++LPsize;
	  ++lrank;
	}
      }

      RS[0][count] = PS_[0][count] = lrank;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("i1=%d, t=%d, rank=%d, PS_[0][%d]=%d\n", i1, t, lrank, count, PS_[0][count]);
*****/

      ++count;

    } /* t */

  } /* i1 */


/***********
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
for(i1=0;i1<count;++i1) 
printf("PS_[0][%d]=%d\n", i1, PS_[0][i1]);
***********/

/***** For debug: print out array *****
print_PRS2_0(PS_[0], 'P');
**************************************/

/***** For debug: print out array *****
print_storage();
**************************************/

}



/******************************************************************************
 *                                                                            *
 *  Function : intermediate();                                                *
 *                                                                            *
 ******************************************************************************/


void intermediate1()
{
  int i;
  int PRS_dim();
  void ex_local_prefix();

  /*--- substep 1 ---*/

  CMMD_node_timer_start(3);

  CMMD_scan_v(PS_[0], PS_[0], CMMD_combiner_add,
              CMMD_upward, CMMD_none, 0, CMMD_exclusive,
              4, ntile[0]);

  CMMD_reduce_v(RS[0], RS[0], CMMD_combiner_add, 4, ntile[0]);

/***
  PRS_dim(0, ntile[0], PS_[0], RS[0]); 
  --- Does not work in cyclic dist (4/27) 
  --- It seems to work, but we need more tests.
      In any case, we'd better use CMMD function for better performance (6/27)
 ***/

  CMMD_node_timer_stop(3);

/***** For debug: print out array *****
print_PRS1(PS_[0], 'P');
print_PRS1(RS[0], 'R');
**************************************/

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n   ===> Substep 1 \n");
*****/

  /*--- substep 2 ---*/
  Psize = RS[0][ntile[0]-1];

  ex_local_prefix(RS[0], ntile[0], 1);

  for (i=0; i<ntile[0]; ++i)
    PS_[0][i] += RS[0][i];

/***** For debug: print out array *****
print_PRS1(PS_[0], 'P');
print_PRS1(RS[0], 'R');
**************************************/

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n    ===> Substep 2 \n");
*****/

  /*--- substep 3 ---*/
  Psize += RS[0][ntile[0]-1];

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n   ===> Substep 3 \n");
*****/


}


/*---------------------------------------------------------------------------*/


void intermediate2()
{
  int t1, i1;
  int i;
  int PRS_dim();
  void ex_local_prefix();

  /*--- substep 1.0 ---*/

  CMMD_node_timer_start(3);

  PRS_dim(0, PRS_size[0], PS_[0], RS[0]); 

  CMMD_node_timer_stop(3);

/***** For debug: print out array *****
print_PRS2_0(PS_[0], 'P');
print_PRS2_0(RS[0], 'R');
**************************************/

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n   ===> Substep 1.0 \n");
*****/

  /*--- substep 2.0 ---*/
  for (t1=0; t1<ntile[1]; ++t1) { /* t1: tile num dim=1 */
    i1 = (t1+1)*width[1] - 1;
    RS[1][t1] = RS[0][ i1*ntile[0] + ntile[0]-1 ];
  }

  ex_local_prefix(RS[0], PRS_size[0], ntile[1]);

  for (i=0; i< PRS_size[0]; ++i)
    PS_[0][i] += RS[0][i];

/***** For debug: print out array *****
print_PRS2_0(PS_[0], 'P');
print_PRS2_0(RS[0], 'R');
**************************************/

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n    ===> Substep 2.0 \n");
*****/

  /*--- substep 3.0 ---*/
  for (t1=0; t1<ntile[1]; ++t1) { /* t1: tile num dim=1 */
    i1 = (t1+1)*width[1] - 1;
    PS_[1][t1] = RS[1][t1] = RS[1][t1] + RS[0][ i1*ntile[0] + ntile[0]-1 ];
  }

/***** For debug: print out array *****
print_PRS2_1(PS_[1], 'P');
print_PRS2_1(RS[1], 'R');
**************************************/

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n   ===> Substep 3.0 \n");
*****/





  /*--- substep 1.1 ---*/

  CMMD_node_timer_start(3);

  PRS_dim(1, PRS_size[1], PS_[1], RS[1]); 

  CMMD_node_timer_stop(3);

/***** For debug: print out array *****
print_PRS2_1(PS_[1], 'P');
print_PRS2_1(RS[1], 'R');
**************************************/

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n   ===> Substep 1.1 \n");
*****/

  /*--- substep 2.1 ---*/
  Psize = RS[1][ntile[1]-1];

  ex_local_prefix(RS[1], ntile[1], 1);

  for (i=0; i<ntile[1]; ++i)
    PS_[1][i] += RS[1][i];

/***** For debug: print out array *****
print_PRS2_1(PS_[1], 'P');
print_PRS2_1(RS[1], 'R');
**************************************/

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n    ===> Substep 2.1 \n");
*****/

  /*--- substep 3.1 ---*/
  Psize += RS[1][ntile[1]-1];

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n   ===> Substep 3.1 \n");
*****/




}



/******************************************************************************
 *                                                                            *
 *  Function : final_step();                                                *
 *                                                                            *
 ******************************************************************************/


void final_step1()
{
  int i, j, k;
  int div;

  Vwidth = Ceiling( Psize, tot_nproc );
  div = Psize / Vwidth;

  if ( my_pnum < div ) 
    V_L = Vwidth;
  else if ( my_pnum == div )
    V_L = Psize % Vwidth;
  else
    V_L = 0;

  /*-------------- Compute rank & Make sendl ------------------*/
  for (i=0; i<LPsize; ++i) {
    storage[i].rank += PS_[0][ storage[i].tnum0 ];
    storage[i].dest = storage[i].rank / Vwidth;
    sendl[storage[i].dest] +=2; /* 2i=data, 2i+1=rank */
  }

}


/*---------------------------------------------------------------------------*/


void final_step2()
{
  int i, j, k;
  int div;

  Vwidth = Ceiling( Psize, tot_nproc );

  div = Psize / Vwidth;

  if ( my_pnum < div ) 
    V_L = Vwidth;
  else if ( my_pnum == div )
    V_L = Psize % Vwidth;
  else
    V_L = 0;

/***** For debug: print out array *****
  print_PRS2_1(PS_[1], 'P');
  print_PRS2_0(PS_[0], 'P');
***************************************/

  /*-------------- PS0 + PS1 -----------------*/

  for (i=0; i<ntile[1]; ++i)
    for (j=i*width[1]; j<(i+1)*width[1]; ++j)
	 for (k=0; k<ntile[0]; ++k)
	   PS_[0][j*ntile[0] + k] += PS_[1][i];

/***** For debug: print out array *****
print_PRS2_0(PS_[0], 'P');
**************************************/

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n   ===> Sum all PS \n");
*****/

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n >>>>>>>>>>>>>> Psize=%d, Nproc=%d, Vwidth=%d \n", 
                Psize, tot_nproc, Vwidth);
*****/

  /*---------------------- Compute rank & Make sendl -------------------------*/

  for (i=0; i<LPsize; ++i) {
    storage[i].rank += PS_[0][ storage[i].lidx[1]*ntile[0] + storage[i].tnum0 ];
    storage[i].dest = storage[i].rank / Vwidth;   
    sendl[storage[i].dest] +=2; /* 2i=data, 2i+1=rank */ 
  }
}



/******************************************************************************
 *                                                                            *
 *  Function : comp_mesg();                                                   *
 *                                                                            *
 ******************************************************************************/


void comp_mesg()
{
  int pn, i;
  int offset;
  void send_sendl_old();
  void send_sendl_new();
  int scount[Procmax];

  /*--- setup send_offset ---*/

  offset = 0;
  for (pn=0; pn<tot_nproc; ++pn) {
    send_offset[pn] = offset;
    offset += sendl[pn];
  }

  /*--- send sendl vector ---*/

  CMMD_node_timer_stop(1);    /* compo */
  CMMD_node_timer_start(8);   /* snd_sndl */

  send_sendl_new();

  CMMD_node_timer_stop(8);    /* snd_sndl */
  CMMD_node_timer_start(1);   /* compo */

  /*--- setup recv_offset ---*/

  offset = 0;
  for (pn=0; pn<tot_nproc; ++pn) {
    recv_offset[pn] = offset;
    offset += recvl[pn];
  }

  /*--- Message coalescing ---*/

  for (pn=0; pn<tot_nproc; ++pn)
    scount[pn] = send_offset[pn];

  for (i=0; i<LPsize; ++i) {
    pn = storage[i].dest;

    if (rank ==1)
      sendB[ scount[pn]++ ]  = A1[ storage[i].lidx[0] ];
    else if (rank == 2)
      sendB[ scount[pn]++ ] =
	A2[ storage[i].lidx[1] ][ storage[i].lidx[0] ];
    else {
      printf("\n Sorry, we can't support rank > 2. Exit! \n\n");
      exit(0);
    };

    sendB[ scount[pn]++ ] = storage[i].rank;

  }

}



/******************************************************************************
 *                                                                            *
 *  Function : decomp_mesg()
 *                                                                            *
 ******************************************************************************/

  
void decomp_mesg()
{

  int pn, i;

  for (pn=0; pn<tot_nproc; ++pn)
    if ( recvl[pn] &&  pn != my_pnum ) 
      for (i=recv_offset[pn]; i<recv_offset[pn]+recvl[pn]; i+=2)
	V[ recvB[i+1] % Vwidth ] = recvB[i];

  /* for my_self */
  if ( sendl[my_pnum] ) {
    for (i=send_offset[my_pnum]; i<send_offset[my_pnum]+sendl[my_pnum]; i+=2)
      V[ sendB[i+1] % Vwidth ] = sendB[i];
  }

/*********
  fcntl( fileno(stdout), F_SETFL, O_APPEND);
  CMMD_fset_io_mode(stdout, CMMD_independent);
  printf("\n===> [PN:%3d] Size = %d  ", my_pnum, V_L);
  for (i=0; i<V_L; ++i) 
  printf("[P:%d, I:%d] %d  ", my_pnum, i, V[i]);
  fflush(stdout);
********/

}
 
void __decomp_mesg__()  /* It does not work well, and inefficient (5/6) */
{

  int pn, i;

  for (i=0; i<2*V_L; i+=2) 
    V[ recvB[i+1]%Vwidth ] = recvB[i];

  /* for my_self */
  if ( sendl[my_pnum] ) {
    for (i=send_offset[my_pnum]; i<send_offset[my_pnum]+sendl[my_pnum]; i+=2)
      V[ sendB[i+1] % Vwidth ] = sendB[i];
  }

/*********
  fcntl( fileno(stdout), F_SETFL, O_APPEND);
  CMMD_fset_io_mode(stdout, CMMD_independent);
  printf("\n===> [PN:%3d] Size = %d  ", my_pnum, V_L);
  for (i=0; i<V_L; ++i) 
  printf("[P:%d, I:%d] %d  ", my_pnum, i, V[i]);
  fflush(stdout);
********/

}



/******************************************************************************
 *                                                                            *
 *  Function : send_sendl()
 *                                                                            *
 ******************************************************************************/


int rpc_ack;

void rpc_handler(address, value)
     int *address;
     int value;
{
   *address = value;
   rpc_ack = true;    
};


void send_sendl_new()
{

  int partner;
  int i;

  CMAML_disable_interrupts();

  for (i=1; i<tot_nproc; ++i) {
    partner = my_pnum ^ i;
    rpc_ack =false;
    CMMD_sync_with_nodes();
    CMAML_rpc(partner, rpc_handler, &recvl[my_pnum], sendl[partner]);
    CMAML_poll_while_lt(&rpc_ack, true); 
  }

  recvl[my_pnum] = sendl[my_pnum];

  CMAML_enable_interrupts();

}


void send_sendl_old()
{

  int partner;
  int i;
  int recv;

  CMMD_sync_with_nodes();

  for (i=1; i<tot_nproc; ++i) {
    partner = my_pnum ^ i;
    CMMD_swap(partner, 
	      &recv,  4,       /* receive */
	      &sendl[partner], 4);     /* send */
    recvl[partner] = recv;
  }

  recvl[my_pnum] = sendl[my_pnum];

}



/******************************************************************************
 *                                                                            *
 *  Function : many_to_many_comm()
 *                                                                            *
 ******************************************************************************/


void many_to_many_comm(comm_alg)
     int comm_alg;
{
  void linear_permutation_am_put();
  void K_linear_permutation();
  void unschedule_cmmd_async();
  void K_unschedule_cmmd_async();
  void linear_permutation_cmmd();
  void two_stage_scheduling_1();  /*  LP_Xor Without Reshuffling */
  void two_stage_scheduling_2();  /*  LP_Xor With Reshuffling */
  void two_stage_scheduling_3();  /*  DS without Reshuffling */
  void two_stage_scheduling_4();  /*  DS with Reshuffling */
  void distributed_scheduling();


  switch (comm_alg) { 
     case 1: unschedule_cmmd_async(); 
             break;

     case 2: linear_permutation_cmmd();
             break;

     case 3: linear_permutation_am_put();
             break;

     case 4: K_unschedule_cmmd_async();
             break;

     case 5: K_linear_permutation();
             break;

     case 6: distributed_scheduling();
             break;

     case 7: two_stage_scheduling_1();
             break;

     case 8: two_stage_scheduling_2(); 
             break;

     case 9: two_stage_scheduling_3(); 
             break;

     case 10: two_stage_scheduling_4(); 
             break;
  }

}



/******************************************************************************
 *                                                                            *
 *  Function : unschedule_cmmd_async()                                        *
 *                                                                            *
 ******************************************************************************/


void unschedule_cmmd_async()
{
  int tp, sp, i;

  CMMD_sync_with_nodes();

  for (tp=0; tp<tot_nproc; ++tp) 
    if ( sendl[tp] && tp != my_pnum )
      CMMD_free_mcb(
		    CMMD_send_async(tp,
				    CMMD_DEFAULT_TAG,
				    &sendB[ send_offset[tp] ],
				    sendl[tp]*msg_size,
				    0,
				    0)
		    );

  for (sp=0; sp<tot_nproc; ++sp) 
    if ( recvl[sp] && sp != my_pnum )
      CMMD_free_mcb(
		    CMMD_receive_async(sp,
				      CMMD_DEFAULT_TAG,
				      &recvB[ recv_offset[sp] ],
				      recvl[sp]*msg_size,
				      0,
				      0)
		    );


  while ( !CMMD_all_msgs_done() );

}

 

/******************************************************************************
 *                                                                            *
 *  Function : linear_permutation_cmmd
 *                                                                            *
 ******************************************************************************/


void linear_permutation_cmmd()
{
  int partner;
  int i;
  int recv;

  CMMD_sync_with_nodes();

  for (i=1; i<tot_nproc; ++i) {
    partner = my_pnum ^ i;
    if ( recvl[partner] != 0 || sendl[partner] != 0 )
      CMMD_swap(partner, 
		&recvB[ recv_offset[partner] ],  /* receive */
		recvl[partner]*msg_size,
		&sendB[ send_offset[partner] ],   /* send */
		sendl[partner]*msg_size);

  }


}


void linear_permutation_cmmd__()
{
  int partner;
  int i;

  CMMD_sync_with_nodes();

  for (i=1; i<tot_nproc; ++i) {
    partner = my_pnum ^ i;

    if (sendl[partner])
      CMMD_free_mcb(
      CMMD_send_async(partner, CMMD_DEFAULT_TAG, 
		      &sendB[ send_offset[partner] ],
		      sendl[partner]*msg_size, 0, 0)
      );

    if (recvl[partner])
       CMMD_receive_block(partner, CMMD_DEFAULT_TAG, 
			  &recvB[ recv_offset[partner] ],
			  recvl[partner]*msg_size);
    CMMD_sync_with_nodes();
  }

  while ( !CMMD_all_msgs_done() );

}



/******************************************************************************
 *                                                                            *
 *  Function : linear_permutation_am_put()                                    *
 *                                                                            *
 ******************************************************************************/


int recv_ack;

void recv_handler()                   
{
   recv_ack = true; 
};


void linear_permutation_am_put()  
{	
  int i, partner;
	
  CMAML_disable_interrupts();

  for (i=1; i<tot_nproc; ++i) {
    partner = my_pnum ^ i;	

    if ( recvl[partner] ) {

      recv_ack = false;

      /* Don't worry about double alignment. Offset is always even. */
      CMAML_set_rport_addr         (rport, &recvB[ recv_offset[partner] ]);
      CMAML_set_rport_handler      (rport, recv_handler);
      CMAML_set_rport_handler_arg2 (rport, null);
      CMAML_set_rport_byte_counter (rport, recvl[partner]*msg_size);
      CMAML_set_rport_total_nbytes (rport, recvl[partner]*msg_size);
    }

    CMMD_sync_with_nodes();

    if (sendl[partner])
     CMAML_scopy(partner,
		 rport,
		 0,
		 0,
		 &sendB[ send_offset[partner] ],
		 sendl[partner]*msg_size,
		 null,
		 null);  

    if (recvl[partner])
      CMAML_poll_while_lt(&recv_ack, true); 
		
  }

  CMAML_enable_interrupts();

/*************** LOCAL COPY ===> ONLY FOR DEBUG ***********

  if ( sendl[my_pnum] ) {
    for (i=0; i<sendl[my_pnum]; ++i)
      recvB[ recv_offset[my_pnum] + i ] = sendB[ send_offset[my_pnum] + i ];
  }

*********************************************************/
}



/******************************************************************************
 *                                                                            *
 *  Function : K_unschedule_cmmd_async()                                      *
 *                                                                            *
 ******************************************************************************/


void  K_unschedule_cmmd_async()
{
  int scount, rcount;

  /*--- initialize scarab ---*/
  PS = tot_nproc;
  CMAML_disable_interrupts();
  Self = CMMD_self_address();

  /*--- Avoid unnecessary local copy when sender = receiver ---*/
  /*--- When decompose message, we directly use sendB       ---*/

  scount = sendl[Self];
  sendl[Self] = 0;

  rcount = recvl[Self];
  recvl[Self] = 0;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* Begin DS_Many_Con\n");
CMMD_fset_io_mode(stdout, CMMD_independent);
*****/

  CMMD_sync_with_nodes();

  ManytoMany_Con(sendB, sendl, send_offset, msg_size, 
                     recvB, recvl, recv_offset, 2, -1, -1);

  /*--- Restore comm vectors ---*/
  sendl[Self] = scount;
  recvl[Self] = rcount;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End DS_Many_Con\n");
*****/

}



/******************************************************************************
 *                                                                            *
 *  Function :  K_linear_permutation()
 *                                                                            *
 ******************************************************************************/


void  K_linear_permutation()
{

  int scount, rcount;

  /*--- initialize scarab ---*/
  PS = tot_nproc;
  CMAML_disable_interrupts();
  Self = CMMD_self_address();

  /*--- Avoid unnecessary local copy when sender = receiver ---*/
  /*--- When decompose message, we directly use sendB       ---*/

  scount = sendl[Self];
  sendl[Self] = 0;

  rcount = recvl[Self];
  recvl[Self] = 0;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* Begin DS_Many_Con\n");
CMMD_fset_io_mode(stdout, CMMD_independent);
*****/

  CMMD_sync_with_nodes();

  ManytoMany_Con(sendB, sendl, send_offset, msg_size, 
                     recvB, recvl, recv_offset, 1, -1, -1);

  CMAML_enable_interrupts();

  /*--- Restore comm vectors ---*/
  sendl[Self] = scount;
  recvl[Self] = rcount;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End DS_Many_Con\n");
*****/

}



/******************************************************************************
 *                                                                            *
 *  Function : distributed_scheduling()                                       *
 *                                                                            *
 ******************************************************************************/


void distributed_scheduling()
{

  int scount, rcount;

  /*--- initialize scarab ---*/
  PS = tot_nproc;
  CMAML_disable_interrupts();
  Self = CMMD_self_address();

  /*--- Avoid unnecessary local copy when sender = receiver ---*/
  /*--- When decompose message, we directly use sendB       ---*/

  scount = sendl[Self];
  sendl[Self] = 0;

  rcount = recvl[Self];
  recvl[Self] = 0;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* Begin DS_Many_Con\n");
CMMD_fset_io_mode(stdout, CMMD_independent);
*****/
  CMMD_sync_with_nodes();

  ManytoMany_Con(sendB, sendl, send_offset, msg_size, 
                     recvB, recvl, recv_offset, 3, -1, -1);

  CMAML_enable_interrupts();

  /*--- Restore comm vectors ---*/
  sendl[Self] = scount;
  recvl[Self] = rcount;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End DS_Many_Con\n");
*****/

}



/******************************************************************************
 *                                                                            *
 *  Function : two_stage_scheduling()                                         *
 *                                                                            *
 ******************************************************************************/


void two_stage_scheduling_1()  /*  LP_Xor Without Reshuffling */
{

  int scount, rcount;

  /*--- initialize scarab ---*/
  PS = tot_nproc;
  CMAML_disable_interrupts();
  Self = CMMD_self_address();

  /*--- Avoid unnecessary local copy when sender = receiver ---*/
  /*--- When decompose message, we directly use sendB       ---*/

  scount = sendl[Self];
  sendl[Self] = 0;

  rcount = recvl[Self];
  recvl[Self] = 0;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* Begin Two_Stage_Many_Con;\n");
CMMD_fset_io_mode(stdout, CMMD_independent);
*****/

  CMMD_sync_with_nodes(); 

  ManytoMany_Con(sendB, sendl, send_offset, msg_size, 
                     recvB, recvl, recv_offset, 4, 1, 1);

  CMAML_enable_interrupts();

  /*--- Restore comm vectors ---*/
  sendl[Self] = scount;
  recvl[Self] = rcount;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End Two_Stage_Many_Con;\n");
*****/

}



void two_stage_scheduling_2() /*  LP_Xor With Reshuffling */
{

  int scount, rcount;

  /*--- initialize scarab ---*/
  PS = tot_nproc;
  CMAML_disable_interrupts();
  Self = CMMD_self_address();

  /*--- Avoid unnecessary local copy when sender = receiver ---*/
  /*--- When decompose message, we directly use sendB       ---*/

  scount = sendl[Self];
  sendl[Self] = 0;

  rcount = recvl[Self];
  recvl[Self] = 0;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* Begin Two_Stage_Many_Con;\n");
CMMD_fset_io_mode(stdout, CMMD_independent);
*****/

  CMMD_sync_with_nodes(); 

  ManytoMany_Con(sendB, sendl, send_offset, msg_size, 
                     recvB, recvl, recv_offset, 4, 2, 2);

  CMAML_enable_interrupts();

  /*--- Restore comm vectors ---*/
  sendl[Self] = scount;
  recvl[Self] = rcount;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End Two_Stage_Many_Con;\n");
*****/

}



void two_stage_scheduling_3()  /* DS without Reshuffling */
{

  int scount, rcount;

  /*--- initialize scarab ---*/
  PS = tot_nproc;
  CMAML_disable_interrupts();
  Self = CMMD_self_address();

  /*--- Avoid unnecessary local copy when sender = receiver ---*/
  /*--- When decompose message, we directly use sendB       ---*/

  scount = sendl[Self];
  sendl[Self] = 0;

  rcount = recvl[Self];
  recvl[Self] = 0;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* Begin Two_Stage_Many_Con;\n");
CMMD_fset_io_mode(stdout, CMMD_independent);
*****/

  CMMD_sync_with_nodes(); 

  ManytoMany_Con(sendB, sendl, send_offset, msg_size, 
                     recvB, recvl, recv_offset, 4, 3, 3);

  CMAML_enable_interrupts();

  /*--- Restore comm vectors ---*/
  sendl[Self] = scount;
  recvl[Self] = rcount;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End Two_Stage_Many_Con;\n");
*****/

}



void two_stage_scheduling_4() /* DS with Reshuffling */
{

  int scount, rcount;

  /*--- initialize scarab ---*/
  PS = tot_nproc;
  CMAML_disable_interrupts();
  Self = CMMD_self_address();

  /*--- Avoid unnecessary local copy when sender = receiver ---*/
  /*--- When decompose message, we directly use sendB       ---*/

  scount = sendl[Self];
  sendl[Self] = 0;

  rcount = recvl[Self];
  recvl[Self] = 0;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* Begin Two_Stage_Many_Con;\n");
CMMD_fset_io_mode(stdout, CMMD_independent);
*****/

  CMMD_sync_with_nodes(); 

  ManytoMany_Con(sendB, sendl, send_offset, msg_size, 
                     recvB, recvl, recv_offset, 4, 4, 4);

  CMAML_enable_interrupts();

  /*--- Restore comm vectors ---*/
  sendl[Self] = scount;
  recvl[Self] = rcount;

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End Two_Stage_Many_Con;\n");
*****/

}




/******************************************************************************
 *                                                                            *
 *  Function : PRS()_dim                                                      *
 *                                                                            *
 *     - Scalable vector PRS on PS_i & RS_i                                   *
 *     - Vector input value is stored in p_result                             *
 *                                                                            *
 ******************************************************************************/


int PRS_dim(dim, size, p_result, r_result)
     int dim,          /* PRS along dimension = dim */
         size,         /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *p_result,    /* exclusive prefix */
         *r_result;    /* reduction */
{
  void PRS_dim_naive();
  void PRS_dim_naive_cmmd();


/***
  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  if (  nproc[dim] < 8 || size < nproc[dim] )
    printf("\n\n*********** NAIVE: dim=%d, size=%d, nproc[dim]=%d \n",
	   dim, size, nproc[dim]);
  else
    printf("\n\n*********** NEW: dim=%d, size=%d, nproc[dim]=%d \n",
	   dim, size, nproc[dim]);
***/

  if ( size < 1 )
    return 0;

  else if ( nproc[dim] < 8 || size < nproc[dim] )
    PRS_dim_naive(dim, size, p_result, r_result);

  else  /* nproc[dim] >= 8 && size >= nproc[dim] */
    PRS_dim_new(dim, size, p_result, r_result);

  return 0;

}



/******************************************************************************
 *                                                                            *
 *  Function : PRS_dim_naive()                                                *
 *                                                                            *
 *     - Scalable vector PRS on PS_i & RS_i using naive algorithm             *  
 *       which is defined Kurmar's book                                       *
 *     - Useful when vector size < number of processors                       *
 *                                                                            *
 ******************************************************************************/


int prs_dim_ack;

void prs_dim_handler()                   
{
   prs_dim_ack = true; 
};

void PRS_dim_naive(dim, size, p_result, r_result)
     int dim,          /* PRS along dimension = dim */
         size,         /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *p_result,    /* exclusive prefix */
         *r_result;    /* reduction */
{

  int in_mesg[Bufmax]; 
  /*** int *in_mesg; ***/

  int partner, partner_grid[Rankmax];
  int i, j;

  /*** in_mesg = (int *)malloc(size * datasize); ***/

  for (j=0; j<size; ++j) {
    /** out_mesg[j] = p_result[j]; **(7/3)**/  /* same as r_result */
    p_result[j] = 0;            /* exclusive prefix */
  }

  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  CMAML_set_rport_total_nbytes (rport, msg_size*size);

  CMMD_sync_with_nodes();

  for (j=0; j<rank; ++j)
    partner_grid[j] = my_grid[j];

  for (i=0; i<dproc[dim]; ++i) {
  
    partner_grid[dim] = my_grid[dim] ^ pow2(i);
    map_to_pnum(partner_grid, &partner);

    prs_dim_ack = false;

    CMAML_set_rport_byte_counter (rport, msg_size*size);

    CMMD_sync_with_nodes();

    CMAML_scopy(partner,
		rport,
		0,
		0,
		r_result, /** out_mesg **(7/3)**/
		msg_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    for (j=0; j<size; ++j)
      r_result[j] += in_mesg[j];

    if (partner_grid[dim] < my_grid[dim])
      for (j=0; j<size; ++j)
	p_result[j] += in_mesg[j];

    partner_grid[dim] = my_grid[dim];

  }

  /** for (j=0; j<size; ++j)
    r_result[j] = out_mesg[j]; **(7/3)**/

  CMAML_enable_interrupts();

  /*** free(in_mesg); ***/

}


/*----------------------------------------------------------------------------*/


void PRS_dim_naive_cmmd(dim, size, p_result, r_result)
     int dim,          /* PRS along dimension = dim */
         size,         /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *p_result,    /* exclusive prefix */
         *r_result;    /* reduction */
{

  int *out_mesg, *in_mesg;

  int partner, partner_grid[Rankmax];
  int i, j;

  out_mesg = (int *)malloc(size * datasize);
  in_mesg = (int *)malloc(size * datasize);

  for (j=0; j<size; ++j)
    p_result[j] = 0;            /* exclusive prefix */

  CMMD_sync_with_nodes();

  for (j=0; j<rank; ++j)
    partner_grid[j] = my_grid[j];

  for (i=0; i<dproc[dim]; ++i) {
  
    partner_grid[dim] = my_grid[dim] ^ pow2(i);
    map_to_pnum(partner_grid, &partner);

    CMMD_swap(partner, 
	      in_mesg, 4*size,       /* receive */
	      out_mesg, 4*size);     /* send */

    for (j=0; j<size; ++j)
      out_mesg[j] += in_mesg[j];

    if (partner_grid[dim] < my_grid[dim])
      for (j=0; j<size; ++j)
	p_result[j] += in_mesg[j];

    partner_grid[dim] = my_grid[dim];

  }

  free(out_mesg);
}

/******************************************************************************
 *                                                                            *
 *          Programmed by Dong-Min Kim                                        *
 *                                                                            *
 ******************************************************************************
 *                                                                            *
 *          Global variables for add_n_store & PRS_dim_new() and              *
 *          Interface with my programs                                        *
 *                                                                            *
 ******************************************************************************/

/*---------- Begin: My modification (7/11/95) -----------*/

#define Max  V_Lmax /* N_Lmax * N_Lmax */
#define IMax 16    /* 16 = log(V_Lmax) */

int    my_id;              /* The ID of the selected dim in the converted ID */
int    dir_nproc;          /* Totol number of processors in the selected dim */

int    partner_grid[Rankmax];   /* The processor ID with which the data 
				   is exchanged */
int    right_grid[Rankmax];     /* The processor ID of right child */
/*int    parent_grid[Rankmax];*/    /* The processor ID of parent */

int    partner_pnum; 
int    right_pnum;
/*int    parent_pnum;*/

/*----------- End: My modification (7/11/95) ------------*/



/*******************************************************************************
 *   Procedure : add_n_store                                                   *
 *   Input     : int start_index, int data_size, int step, int *v_1, int *v_2  *
 *               int *v_3[]                                                    *
 *   Output    : None                                                          *
 *   Function  : Add the two arrays from the start_index by data_size, and     *
 *                store the right child's data.                                *
 *******************************************************************************/


add_n_store(start_index, data_size, step, v_1, v_2, v_3)
     int start_index,
         data_size,
         step,
         *v_1, *v_2, v_3[IMax][Max];
{
  int i;


  if ( right_pnum == partner_pnum ) {
    for ( i = start_index ; i < start_index + data_size ; i++ ) {
      v_1[i] += v_2[i];
      v_3[step][i] = v_2[i];
    }
  }
  else {
    for ( i = start_index ; i < start_index + data_size ; i++ ) {
      v_3[step][i] = v_1[i];
      v_1[i] += v_2[i];
    }
  }

}


show(id, start, size, vector)
     int id, start, size, *vector;
{
  int pn;
  int i;


  CMMD_fset_io_mode(stdout, CMMD_independent);
  for ( pn = 0 ; pn < tot_nproc ; pn++ ) {
    if ( id == pn ) {
      printf("***********************************************\n");
      printf("Processor %2d =>  ", id);

      for ( i = 0 ; i < rank ; i++ )
	printf("%d ", my_grid[i]);
      printf("\n");
	  
      for ( i = start ; i < start+size ; i++ )
	printf("%d  :  %3d\n", i, vector[i]);

      printf("***********************************************\n");
      fflush(stdout);
    }

    CMMD_sync_with_nodes();
  }
}



/******************************************************************************
 *                                                                            *
 *  Function : PRS_dim_new()                                                  *
 *                                                                            *
 *     - Scalable vector PRS on PS_i & RS_i using advanced algorithm          *  
 *     - Useful when vector size is large                                     *
 *                                                                            *
 ******************************************************************************/


PRS_dim_new(direction, vector_size, input_v, p_result, r_result)
     int direction;
     int vector_size,  /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *input_v,     /* input array */
         *p_result,    /* prefix sum output array */
         *r_result;    /* reduction output array */
{
  int    save[Max];   /* Only for exclusive */
  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  int    store_recv; /* the receive point after the sum */
  int    store_size; /* the data_size after the sum */
  int    Right[IMax][Max];   /* Sums from right child process by step.
			       1st dimension is for step, and
			       2nd dimension is for the sums
			       from the right child process.          */
  int    in_mesg[Max];      /* received data */

/*  show(my_pnum, 0, vector_size, p_result);*/


/*---------- Begin; My modification (7/11/95) -----------*/

  my_id = my_grid[direction]; 

  dir_nproc  = nproc[direction];

  /* setup the partner, right child and the parent */
  for ( i = 0 ; i < rank ; i++ ) {
    partner_grid[i] = my_grid[i];
    right_grid[i] = my_grid[i];
    /*parent_grid[i] = my_grid[i];*/
   }

/*----------- End; My modification (7/11/95) ------------*/

  /** For Active Message Passing **/
  CMAML_disable_interrupts();
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  /********************************/

  CMMD_sync_with_nodes();

  if (input_v == p_result)
    for ( i = 0 ; i < vector_size ; i++ )   /* Only for exclusive */
      save[i] = input_v[i];
  else
    for ( i = 0 ; i < vector_size ; i++ ) 
      p_result[i] = input_v[i];


/*----------- End; My modification (7/11/95) ------------*/

  size = sizeof(int);

  data_size = vector_size;
  start_recv = 0;
  digit = Ilog2(dir_nproc);

/* ===============
        Sum
   =============== */

  for ( step = 0 ; step < digit ; step++ ) {

    data_size = data_size / 2;
    start_send = start_recv;
    check = my_id & pow2(step);
    if ( check == 0 )
      start_send += data_size;
    else
      start_recv += data_size;

    /* decide the partner and the right child */
    partner_pnum = my_id ^ pow2(step);
    partner_grid[direction] = partner_pnum;
    map_to_pnum(partner_grid, &partner_pnum);

    right_pnum   = my_id | pow2(step);
    right_grid[direction] = right_pnum;
    map_to_pnum(right_grid, &right_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_addr         (rport, &in_mesg[start_recv]);
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
                rport,
                0,
                0,
                &p_result[start_send],
                data_size*size,
                null,
                null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    add_n_store(start_recv, data_size, step, p_result, in_mesg, Right);
  }

  store_recv = start_recv;
  store_size = data_size;

  /* copy the p_result to r_result for the reduction */
  for ( i = 0 ; i < vector_size ; i++ )
    r_result[i] = p_result[i];


/* ===============
     PREFIX SUM
   =============== */

/* Distribution */

  start_send = start_recv;
  for ( step = digit - 1 ; step >= 0 ; step-- ) {

    check = my_id & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    /* decide the partner and the right child */
    partner_pnum = my_id ^ pow2(step);
    partner_grid[direction] = partner_pnum;
    map_to_pnum(partner_grid, &partner_pnum);

    right_pnum   = my_id | pow2(step);
    right_grid[direction] = right_pnum;
    map_to_pnum(right_grid, &right_pnum);

    if ( partner_pnum == right_pnum ) {

      /** For Active Message Passing **/

      prs_dim_ack = false;

      /* receive */
      CMAML_set_rport_addr         (rport, &p_result[start_recv]);
      CMAML_set_rport_byte_counter (rport, data_size*size);
      CMAML_set_rport_total_nbytes (rport, data_size*size);

      CMMD_sync_with_nodes();

      /* send */
      CMAML_scopy(partner_pnum,
		  rport,
		  0,
		  0,
		  &p_result[start_send],
		  data_size*size,
		  null,
		  null);  

      CMAML_poll_while_lt(&prs_dim_ack, true); 

      /********************************/

      for ( i = start_send ; i < start_send + data_size ; i++ )
	p_result[i] -= Right[step][i];

    }
    else {
      for ( i = start_send ; i < start_send + data_size ; i++ )
	in_mesg[i] = p_result[i] - Right[step][i];

      /** For Active Message Passing **/

      prs_dim_ack = false;

      /* receive */
      CMAML_set_rport_addr         (rport, &p_result[start_recv]);
      CMAML_set_rport_byte_counter (rport, data_size*size);
      CMAML_set_rport_total_nbytes (rport, data_size*size);

      CMMD_sync_with_nodes();

      /* send */
      CMAML_scopy(partner_pnum,
		  rport,
		  0,
		  0,
		  &in_mesg[start_send],
		  data_size*size,
		  null,
		  null);  

      CMAML_poll_while_lt(&prs_dim_ack, true); 

      /********************************/
    }

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

/*---------- Begin: My modification (7/11/95) -----------*/

  /*show(my_pnum, 0, vector_size, p_result);*/

/* For exclusive prefix */
  if (input_v == p_result)
    for ( i = 0 ; i < vector_size ; i++ )
      p_result[i] -= save[i];
  else
    for ( i = 0 ; i < vector_size ; i++ )
      p_result[i] -= input_v[i];

/*----------- End: My modification (7/11/95) ------------*/


/* ===============
     REDUCTION
   =============== */

/* all-to-all BC */

  data_size  = store_size;
  start_recv = store_recv;
  start_send = start_recv;

  for ( step = digit - 1 ; step >= 0 ; step-- ) {

    check = my_id & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    /* decide the partner */
    partner_pnum = my_id ^ pow2(step);
    partner_grid[direction] = partner_pnum;
    map_to_pnum(partner_grid, &partner_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_addr         (rport, &r_result[start_recv]);
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
		rport,
		0,
		0,
		&r_result[start_send],
		data_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }


  /*show(my_pnum, 0, vector_size, r_result);*/
}

 

/******************************************************************************
 *                                                                            *
 *  Function : ex_local_prefix()                                              *
 *                                                                            *
 *     - Exclusive local prefix on RS_i                                       *
 *     - Assumption: Vector size \ # of segments                              *
 *                                                                            *
 ******************************************************************************/


void ex_local_prefix(a, size, num_seg)
     int *a,       /* input array = RS_i */
         size,     /* array size = L_(i+1) * T_i */
         num_seg;  /* # of segments = T_(i+1) */
{
  int seg_size;    /* segment_size = size / num_seg                 */
                   /*               = ( L_(i+1) * T_i ) / T_(i+1) ) */
  int sum;
  int pass;
  int i, j;

  seg_size = size / num_seg;

  for (i=0; i<num_seg; ++i) {
    sum = 0;
    for (j = i*seg_size; j < (i+1)*seg_size; ++j) {
      pass = a[j];
      a[j] = sum;
      sum += pass;
    }
  }
}



/******************************************************************************
 *                                                                            *
 *  Function : main()                                                         *
 *                                                                            *
 ******************************************************************************/


#define Default 3

main(argc, argv)
     int argc;
     char *argv[];
{

  int i;
  int len, digit;
  int comm_alg;

  double total,     /* ranking + redist */
         comp_tot,  /* rank_comp + compo */

           ranking,   /* Ranking time = t_init + t_lscan + t_inter + t_final */
           ranking2,  /* alternative ranking: Timer 0 */

             t_prs,     /* time for PRS: Timer 3 */
             rank_comp, /* time for computation in ranking */

             t_init,    /* time for initial step: Timer 4 */
             t_lscan,   /* time for local scan step: Timer 5 */
             t_inter,   /* time for intermediate step: Timer 6 */
             t_final,   /* time for final step: Timer 7 */

           redist,    /* redistribution = comm_tot + compo */

             compo,     /* compo + decompo: Timer 1 */

             comm_tot,  /* total comm = comm + snd_sndl */
               comm,      /* many-to-many communication: Timer 2 */
               snd_sndl,  /* send_sendl: Timer 8 */

         maxx, minn, ave;


  /*--------------------- Process the command line ----------------------------*/

  if (argc > 3 || argc < 2) {
    printf("\n Improper command line. Exit! \n");
    printf("\n   [Usuage] p mask-file    algorithm");
    printf("\n         or p mask-density algorithm\n\n");
    exit(0);
  };

  len = strlen(argv[1]);

  digit = true;
  for(i=0; i<len; i++) 
    if(!isdigit(argv[1][i])) digit = false;

  if (argc == 3)
    comm_alg = atoi(argv[2]);
  else 
    comm_alg = Default;

  if ( comm_alg < 1 || comm_alg > 10 ) {
    CMMD_fset_io_mode(stdout, CMMD_sync_bc);
    printf("\n************************************************************");
    printf("\n   1: unschedule_cmmd_async ");
    printf("\n   2: linear_permutation_cmmd ");
    printf("\n   3: linear_permutation_am_put ");
    printf("\n   4: Scarab unschedule_async ");
    printf("\n   5: Scarab linear_permutation ");
    printf("\n   6: distributed_scheduling ");
    printf("\n   7: two_stage_scheduling using LP without reshuffling");
    printf("\n   8: two_stage_scheduling using LP with reshuffling");
    printf("\n   9: two_stage_scheduling using DS without reshuffling");
    printf("\n  10: two_stage_scheduling using DS with reshuffling");
    printf("\n************************************************************\n\n");
    exit(0);
  }

  /*---------------------------- Read Input Data  -----------------------------*/

  read_input();

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End read_input();\n");
*****/

  CMMD_node_timer_clear(0);  /* ranking */
  CMMD_node_timer_clear(1);  /* compo/decomp */
  CMMD_node_timer_clear(2);  /* send_sendl & many-to-many comm */
  CMMD_node_timer_clear(3);  /* prs */
  CMMD_node_timer_clear(4);  /* initial */
  CMMD_node_timer_clear(5);  /* local scan */
  CMMD_node_timer_clear(6);  /* intermediate */
  CMMD_node_timer_clear(7);  /* final step */
  CMMD_node_timer_clear(8);  /* send_sendl */

  /*------------------------------ Initialize ---------------------------------*/
  
  CMMD_sync_with_nodes();

  CMMD_node_timer_start(0);   /* ranking */
  CMMD_node_timer_start(4);   /* initial */

  initialize();

  CMMD_node_timer_stop(0);    /* ranking */
  CMMD_node_timer_stop(4);    /* initial */

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End initialize();\n");
*****/

/***** For debug: make input array *****/
  if ( rank == 1 )
    make_array1();
  else if ( rank == 2 )
    make_array2();
/**************************************/


  /*---------------------------- Make Mask Array  -----------------------------*/

  if (rank==1) {
    if (digit)
      random_mask1(atoi(argv[1]));
    else
      read_mask1(argv[1]);
  }
  else if (rank==2) {
    if (digit)
      random_mask2(atoi(argv[1]));
    else
      read_mask2(argv[1]);
  }
  else {
    printf("\n Sorry, we can't support rank > 2. Exit! \n\n");
    exit(0);
  };

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End make mask array();\n");
*****/

  /*-------------------------- Step 1: Local scan ----------------------------*/

  CMMD_node_timer_start(0);   /* ranking */
 
  CMMD_node_timer_start(5);   /* local scan */

  if (rank == 1)
    local_scan1();
  else
    local_scan2();

/* for only exact timing - i.e., no problem without this barrier */
CMMD_sync_with_nodes(); 

  CMMD_node_timer_stop(5);    /* local scan */

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End local_scan;\n");
*****/

  /*------------------------- Step 2: Intermediate ---------------------------*/

  CMMD_node_timer_start(6);   /* intermediate */

  if (rank == 1)
    intermediate1();
  else
    intermediate2();

  CMMD_node_timer_stop(6);    /* intermediate */


/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End intermediate();\n");
*****/


  /*------------------------- Check if Mask is empty -------------------------*/

  CMMD_node_timer_stop(0); /* ranking */

  if ( ! CMMD_reduce_int(LPsize, CMMD_combiner_add) ) {
    V_L = Psize = 0;
    CMMD_fset_io_mode(stdout, CMMD_sync_bc);
    printf("\n************ All values of mask array are false. Exit!\n\n");
    exit(0);
  }

  CMMD_node_timer_start(0); /* ranking */

  /*-------------------------- Step 3: Final step ----------------------------*/

  CMMD_node_timer_start(7);   /* final step */

  if (rank == 1)
    final_step1();
  else
    final_step2();

/* for only exact timing - i.e., no problem without this barrier */
CMMD_sync_with_nodes(); 

  CMMD_node_timer_stop(7);   /* final step */


/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End final_step();\n");
*****/

  CMMD_node_timer_stop(0); /* ranking */

  /*------------------------ Communication -----------------------------------*/

  CMMD_node_timer_start(1);  /* compo */

  comp_mesg();

/* for only exact timing - i.e., no problem without this barrier */
CMMD_sync_with_nodes(); 
  CMMD_node_timer_stop(1);  /* compo */

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End comp_mesg();\n");
*****/

  CMMD_node_timer_start(2);   /* comm */

  many_to_many_comm(comm_alg);

/* for only exact timing - i.e., no problem without this barrier */
CMMD_sync_with_nodes();

  CMMD_node_timer_stop(2);    /* comm */

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End many_to_many_comm();\n");
*****/

  CMMD_node_timer_start(1);  /* decompo */

  decomp_mesg();

/* for only exact timing - i.e., no problem without this barrier */
/*CMMD_sync_with_nodes();*/ 
  CMMD_node_timer_stop(1);  /* decompo */

/*****
CMMD_fset_io_mode(stdout, CMMD_sync_bc);
printf("\n******* End decomp_mesg();\n");
*****/


  /*------------------------ Timiing data -----------------------------------*/

  ranking2= CMMD_node_timer_busy(0) * 1000.;  /* milisec */
  compo   = CMMD_node_timer_busy(1) * 1000.;  /* milisec */
  comm    = CMMD_node_timer_busy(2) * 1000.;  /* milisec */
  t_prs   = CMMD_node_timer_busy(3) * 1000.;  /* milisec */
  t_init  = CMMD_node_timer_busy(4) * 1000.;  /* milisec */
  t_lscan = CMMD_node_timer_busy(5) * 1000.;  /* milisec */
  t_inter = CMMD_node_timer_busy(6) * 1000.;  /* milisec */
  t_final = CMMD_node_timer_busy(7) * 1000.;  /* milisec */
  snd_sndl= CMMD_node_timer_busy(8) * 1000.;  /* milisec */


  ranking   = t_init + t_lscan +  t_inter +  t_final;
  comm_tot  = comm + snd_sndl;
  redist    = compo + comm_tot;
  total     = ranking + redist;
  rank_comp = ranking - t_prs;
  comp_tot  = rank_comp + compo;

  maxx  = CMMD_reduce_double(total, CMMD_combiner_dmax);
  minn  = CMMD_reduce_double(total, CMMD_combiner_dmin);
  ave  = CMMD_reduce_double(total, CMMD_combiner_dadd) / tot_nproc;



  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
/*****
if (digit)
  printf("[G: %d, %d] [L: %d, %d] [P: %d, %d] [W: %d, %d] [M=%d]\n",
          N[0], N[1], L[0], L[1], nproc[0], nproc[1], width[0], width[1],atoi(argv[1]));
else
  printf("[G: %d, %d] [L: %d, %d] [P: %d, %d] [W: %d, %d] [M=LT]\n",
          N[0], N[1], L[0], L[1], nproc[0], nproc[1], width[0], width[1]);


  printf(" Total |  Rank   Lcomp initial    lsca int_comp   fina int_prs |  Dist   Compo    Comm\n");
  printf("%6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  \n", 
	 total,
	 ranking, rank_comp, t_init, t_lscan, t_inter-t_prs, t_final,
         t_prs,
	 redist, compo, comm_tot);


  printf("Wid  Total | Lcomp |  MtoM |   PRS |Tot_comm\n");
  printf("%3d %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  \n", 
                 width[0], total, comp_tot, comm_tot, t_prs, comm_tot+t_prs);

*****/

/*  printf("Wid  Total | Lcomp |  MtoM |   PRS |T_comm ##  Total |  Rank   Lcomp initial    lsca int_comp   fina int_prs |  Dist   Compo    Comm\n"); */

if (digit)
  printf("%5d %6.2f  %6.2f  %6.2f  %6.2f  %6.2f ## %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  [G: %d, %d] [L: %d, %d] [P: %d, %d] [W: %d, %d] [M=%d] [C=%d]\n",
width[0], total, comp_tot, comm_tot, t_prs, comm_tot+t_prs,
	 total,
	 ranking, rank_comp, t_init, t_lscan, t_inter-t_prs, t_final,
         t_prs,
	 redist, compo, comm_tot,
N[0], N[1], L[0], L[1], nproc[0], nproc[1], width[0], width[1],atoi(argv[1]), comm_alg);

else

  printf("%5d %6.2f  %6.2f  %6.2f  %6.2f  %6.2f ## %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  [G: %d, %d] [L: %d, %d] [P: %d, %d] [W: %d, %d] [M=LT] [C=%d]\n",
width[0], total, comp_tot, comm_tot, t_prs, comm_tot+t_prs,
	 total,
	 ranking, rank_comp, t_init, t_lscan, t_inter-t_prs, t_final,
         t_prs,
	 redist, compo, comm_tot,
N[0], N[1], L[0], L[1], nproc[0], nproc[1], width[0], width[1], comm_alg);

}

